@@ -243,6 +243,7 @@ class Proxy(object):
         util.get_serializer(config.SERIALIZER)  # assert that the configured serializer is available
         self.__async = False
         current_context.annotations = {}
+        current_context.response_annotations = {}
 
     @property
     def _pyroHmacKey(self):
@@ -625,11 +626,10 @@ class Proxy(object):
         """returns a helper class that lets you create batched method calls on the proxy"""
         return _BatchProxyAdapter(self)
 
-    def _pyroAsync(self):
-        """returns an async version of the proxy so you can do asynchronous method calls"""
-        asyncproxy = self.__copy__()
-        asyncproxy.__async = True
-        return asyncproxy
+    def _pyroAsync(self, async=True):
+        """turns the proxy into async mode so you can do asynchronous method calls"""
+        self.__async = async
+        return self
 
     def _pyroInvokeBatch(self, calls, oneway=False):
         flags = message.FLAGS_BATCH
@@ -692,10 +692,18 @@ class _StreamResultIterator(object):
         return self.__next__()
 
     def __next__(self):
+        if self.proxy is None:
+            raise StopIteration
         if self.proxy._pyroConnection is None:
             raise errors.ConnectionClosedError("the proxy for this stream result has been closed")
         self.pyroseq += 1
-        return self.proxy._pyroInvoke("get_next_stream_item", [self.streamId], {}, objectId=constants.DAEMON_NAME)
+        try:
+            return self.proxy._pyroInvoke("get_next_stream_item", [self.streamId], {}, objectId=constants.DAEMON_NAME)
+        except (StopIteration, GeneratorExit):
+            # when the iterator is exhausted, the proxy is removed to avoid unneeded close_stream calls later
+            # (the server has closed its part of the stream by itself already)
+            self.proxy = None
+            raise
 
     def __del__(self):
         self.close()
@@ -711,7 +719,6 @@ class _StreamResultIterator(object):
                 # it decides to gc old iterator objects *during a new call on the proxy*.
                 # If we use the same proxy and do a call in between, the other call on the proxy will get an out of sync seq and crash!
                 # We create a temporary second proxy to call close_stream on. This is inefficient, but avoids the problem.
-                # or use a HACK to decrease the proxy's sequence number by one so it will be unchanged after this call.
                 try:
                     with self.proxy.__copy__() as closingProxy:
                         closingProxy._pyroInvoke("close_stream", [self.streamId], {}, flags=message.FLAGS_ONEWAY, objectId=constants.DAEMON_NAME)
@@ -807,6 +814,17 @@ class _AsyncRemoteMethod(object):
                 # use a copy of the proxy otherwise calls would still be done in sequence,
                 # and use contextmanager to close the proxy after we're done
                 with self.__proxy.__copy__() as proxy:
+                    delay = 0.1 + random.random()/5
+                    while not proxy._pyroConnection:
+                        try:
+                            proxy._pyroBind()
+                        except errors.CommunicationError as x:
+                            if "no free workers" not in str(x):
+                                raise
+                            time.sleep(delay)   # wait a bit until a worker might be available again
+                            delay += 0.4 + random.random()/2
+                            if 0 < config.COMMTIMEOUT/2 < delay:
+                                raise
                     value = proxy._pyroInvoke(self.__name, args, kwargs)
                 asyncresult.value = value
                 return
@@ -827,9 +845,9 @@ def batch(proxy):
     return proxy._pyroBatch()
 
 
-def async(proxy):
-    """convenience method to get an async proxy"""
-    return proxy._pyroAsync()
+def async(proxy, async=True):
+    """convenience method to set proxy to async mode"""
+    return proxy._pyroAsync(async)
 
 
 def pyroObjectToAutoProxy(obj):
@@ -969,6 +987,7 @@ class DaemonObject(object):
         try:
             return next(stream)
         except Exception:
+            # in case of error (or StopIteration!) the stream is removed
             del self.daemon.streaming_responses[streamId]
             raise
 
@@ -1313,7 +1332,7 @@ class Daemon(object):
                 request_seq = msg.seq
                 request_serializer_id = msg.serializer_id
             if xt is not errors.ConnectionClosedError:
-                if xt is not StopIteration:
+                if xt not in (StopIteration, GeneratorExit):
                     log.debug("Exception occurred while handling request: %r", xv)
                 if not request_flags & message.FLAGS_ONEWAY:
                     if isinstance(xv, errors.SerializeError) or not isinstance(xv, errors.CommunicationError):
